#!/usr/bin/perl -w
#
# BootStrip.pl v1.2.2
# Last modified April 22, 2005 11:49
# (c) Olaf R.P. Bininda-Emonds
#
# Input:
#   1) A nexus-formatted treefile containing one or more tree descriptions
#   2) A PAUP* logfile of a bootstrap analysis containing a table of group
#      frequencies
#
# Generates:
#   A nexus-formatted treefile (BootStrip.tre) containing the trees in (1)
#   with bootstrap frequencies derived from (2) represented as branch lengths, node labels, or both
#
# Usage: BootStrip.pl [-b] [-d<number>] [-f<l|n|b>] [-l<filename>] [-r<rootword>] [-t<filename>] [-v] [-h]
#   options: -b = process all tree blocks
#			 -d<number> = number of decimal places to print bootstrap frequencies
#            -f<l|n|b> = include bootstrap frequencies as branch lengths (l - default), node labels (n), or both (b)
#            -l<filename> = log file of bootstrap analysis
#                           (default = bootstrap.log)
#			 -r<rootword> = common root word for -l (rootword_bootlog.txt) and -t (rootword.tre) flags
#            -t<filename> = file containing trees in NEXUS format
#                           (default = treefile.tre)
#            -v = verbose output
#            -h = print this message and quit

use strict;

# Set default parameters
	# I/O parameters
		my $block = 0;
			my $treeBlock = 0;
		my $rootWord;
			my $treeFile = "treefile.tre";
			my $outFile = "treeFile_bootstrip.tre";
			my $logfile = "bootstrap.log";
		
	# Tree parameters
		my (%treeName, %treeDescription, %rootingStatus);
		my $unrootedFlag = 0;
		my $defaultRooting = 0;	# Default is for trees to be unrooted
		my $SupportFlag = 0;
		my @allTaxa;
		
	# Printing parameters
		my $decimalPlace = 3;
		my $labelType = "branch";

	# Miscellaneous parameters
		my $verbose = 0;
		my $debug = 0;

for (my $i = 0; $i <= $#ARGV; $i++)
	{
	if ($ARGV[$i]  eq "-b")
		{
		$block = 1;
		}
	elsif ($ARGV[$i] =~ /-d([\d]+)/)
		{
		$decimalPlace = $1;
		}
	elsif ($ARGV[$i] eq "-fl")
		{
		$labelType = "branch";
		}
	elsif ($ARGV[$i] eq "-fn")
		{
		$labelType = "node";
		}
	elsif ($ARGV[$i] eq "-fb")
		{
		$labelType = "both";
		}
	elsif ($ARGV[$i] =~ /-l([\w+\.?\w+?]*)/)
		{
		$logfile = $1;
		}
	elsif ($ARGV[$i] =~ /-r/)
		{
		($rootWord = $ARGV[$i]) =~ s/^-r//;
		$treeFile = $rootWord . ".tre";
		$outFile = $rootWord . "_bootstrip.tre";
		$logfile = $rootWord . "_bootlog.txt";		
		}
	elsif ($ARGV[$i] =~ /-t([\w+\.?\w+?]*)/)
		{
		$treeFile = $1;
		($outFile = $treeFile) =~ s/\.\w+$//;
		$outFile .= "_bootstrip.tre";
		}
	elsif ($ARGV[$i] eq "-v")
		{
		$verbose = 1;
		}
	elsif ($ARGV[$i] eq "-x")
		{
		$debug = $verbose = 1;
		}
	elsif ($ARGV[$i] eq "-h")
		{
		print "Usage: BootStrip.pl [-b] [-d] [-f<l|n|b>] [-l<filename>] [-r<rootword>] [-t<filename>] [-v] [-h]\n";
		print "Options: -b = process all tree blocks\n";

		print "         -f<l|n|b> = include bootstrap frequencies as branch lengths (l - default), node labels (n), or both (b)\n";
		print "         -d<number> = number of decimal places to print bootstrap frequencies\n";
		print "                      (default = 3)\n";
		print "         -l<filename> = log file of bootstrap analysis\n";
		print "                        (default = bootstrap.log)\n";
		print "         -r<rootword> = common root word for -l (rootword_bootlog.txt) and -t (rootword.tre) flags\n";
		print "         -t<filename> = file containing trees in NEXUS format\n";
		print "                        (default = treefile.tre)\n";
		print "         -v = verbose output\n";
		print "         -h = print this message and quit\n";
		exit(0);
		}
	else
		{
		print "Don't understand argument: $ARGV[$i]\n";
		print "Usage: BootStrip.pl [-b] [-d] [-f<l|n|b>] [-l<filename>] [-r<rootword>] [-t<filename>] [-v] [-h]\n";
		exit(1); 
		}
	}
	
my $decimalPrint = "%.$decimalPlace"."f";

print "\nReading in tree descriptions ...\n" if ($verbose);
	my @allTrees = treeRead($treeFile);
	die "No trees present in $treeFile" if (not @allTrees);

# Read in bootstrap frequencies and taxon IDs
	setLineBreak($logfile);
	open (LOG, "<$logfile") or die "Cannot open bootstrap logfile $logfile\n";
		my (%paupID, %dotPlot, @dotPlot, %bootFreq, $numReps, $minFreq);
		my $bootReadFlag = 0;
		my $dotReadFlag = 0;
		my $lineCount = 0;
		my $numBipartitions = 0;
	
		while (<LOG>)
			{
			chomp;
			my $treeLine = $_;
				$treeLine =~ s/^\s+//;
				$treeLine =~ s/\s+$//;
			next unless ($treeLine);
#				print "$treeLine\n";
			
			# Set flags to only read in appropriate lines
				$bootReadFlag = 1 if ($treeLine =~ /Bootstrap 50%/i or $treeLine =~ /^50% Majority-rule consensus/i);
				if ($treeLine =~ /\(relative\) frequency less than ([\d]+)/i)
					{
					$minFreq = $1;
					$bootReadFlag = $dotReadFlag = 0;
					}
				if ($treeLine =~ /^Processing of file/)
					{
					$minFreq = 0.1 if (not defined $minFreq);
					$bootReadFlag = $dotReadFlag = 0;
					}
				next unless ($bootReadFlag);
	
			# Get number of bootstrap replicates
				if ($treeLine =~ /Number of bootstrap replicates/)
					{
					(my $header, $numReps) = split(/=/, $treeLine);
					$numReps =~ s/\s+//g;
					}
				if ($treeLine =~ /^50% Majority-rule consensus of (\d+) trees/)
					{
					$numReps = $1;
					}

			# Read in taxon IDs	
				if ($treeLine =~ /\(\d+\)/)
					{
					my ($header, $fullTaxon) = split(/\-+\s+/, $treeLine);
					my ($name, $ID) = split(/\(/, $fullTaxon);
						$name =~ s/\s+/_/g;
						$ID =~ s/\)\s*$//;
					$paupID{$name} = $ID;			
					}
			
			# Read in group frequency table
				$dotReadFlag = 1 if ($treeLine =~ /\(bootstrap support values\)\:/ or $treeLine =~ /^Bipartitions found in one or more trees/);
				
				$lineCount = 0 if ($treeLine =~ /^-------/ and $dotReadFlag);
					$lineCount++ if ($dotReadFlag and ($treeLine =~ /\./ or $treeLine =~ /\*/));
					$numBipartitions++ if ($lineCount > $numBipartitions);
					
				$dotPlot{$lineCount} .= $treeLine if ($dotReadFlag and ($treeLine =~ /\./ or $treeLine =~ /\*/));
			}
	close LOG;
	
	die "No taxon IDs present in $logfile" if (not %paupID);
	die "No group frequencies present in $logfile" if (not $numBipartitions);
	
	# Process dot plots to derive bootstrap frequencies
		for (my $plotNum = 1; $plotNum <= $numBipartitions; $plotNum++)
			{
#			print "$plotNum: $dotPlot{$plotNum}\n" if ($debug);
			my (@plotLine) = split(/\s+/, $dotPlot{$plotNum});	# plotLine columns are dotPlot, raw frequency, and (majRule only) percentage

			push @dotPlot, $plotLine[0];
			$bootFreq{$plotLine[0]} = sprintf("$decimalPrint", $plotLine[1] / $numReps * 100);
			}

# Get taxon list and process
	my @taxonList = duplStrip(@allTaxa);
#	my %dupl;	#Remove duplicate taxa
#	my @AllTaxa = grep(!$dupl{$_}++, @taxonList);

	@taxonList = sort { $paupID{$a} <=> $paupID{$b} } keys %paupID;
		my $nullDot = "." x scalar(@taxonList);
		my $fullDot = "*" x scalar(@taxonList);
	
	# Renumber paup IDs for all taxa to remove any gaps (due to taxon deletion)
		for (my $i = 0; $i < scalar(@taxonList); $i++)
			{
			$paupID{$taxonList[$i]} = $i + 1;
			}

print "\nParsing tree descriptions and assigning bootstrap frequencies ...\n" if ($verbose);
	my $treeCount = 0;
	open (OUT, ">$outFile") or die "Cannot open output file $outFile\n";
		print OUT "#NEXUS\n\nbegin trees;\n";
		foreach my $treeID (@allTrees)	# Get structure of each tree
			{
			my $tree = $treeDescription{$treeID};
			$treeCount++;
			my $treeGrow;
			my ($offset, @ParentClade, @open, @start, @stop, $OldCurrent, @CladeDescrip);
			print "\nTree: $treeCount\n" if ($verbose);
			my ($position, $CladeID, $CurrentClade) = (0, 0, 0);
			my %present;
			while ($position < length($tree))
				{
				$offset = 1;
				if (substr($tree, $position, 1) eq "(")		# New clade; go one level deeper
					{
					$CladeID++;
					$ParentClade[$CladeID] = $CurrentClade;
					$CurrentClade = $CladeID;
					$open[$CurrentClade] = 1;
					$start[$CladeID] = $position;
					$treeGrow .= "(";
					}
				elsif (substr($tree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
					{
					$open[$CurrentClade] = 0;
					$OldCurrent = $CurrentClade;
					$CurrentClade = $ParentClade[$OldCurrent];
					$stop[$OldCurrent] = $position;
					$CladeDescrip[$OldCurrent] = substr($tree, $start[$OldCurrent], $stop[$OldCurrent]-$start[$OldCurrent]+1);
					$CladeDescrip[$OldCurrent] =~ s/:\d+\.?\d*e?\-?\d*//g;	# Remove branch length information
					my $presentDot = $nullDot;
					my $absentDot = $fullDot;
					print "\t$OldCurrent\t$CladeDescrip[$OldCurrent]\n" if ($verbose);
					if ($CurrentClade == 0)
						{
						$treeGrow .= ")";
						$treeGrow .= "'n/a'" if ($labelType eq "both" or $labelType eq "node");
						$treeGrow .= ":0" if ($labelType eq "both" or $labelType eq "branch");
						print "\t\tBootFreq: n/a\n" if ($verbose);
						}
					else
						{
						$treeGrow .= ")";
						my $cladeBF;
						foreach my $taxon (@taxonList)	# Derive dot plot for that clade
							{
							if ($present{$taxon}{$OldCurrent})
								{
								substr($presentDot, $paupID{$taxon}-1, 1) = "*";
								substr($absentDot, $paupID{$taxon}-1, 1) = ".";
								}
							}
						print "\t\t$presentDot\n\t\t$absentDot\n" if ($verbose);

						foreach my $entry (@dotPlot)
							{
							if (($presentDot eq $entry) or ($absentDot eq $entry))
								{
								$cladeBF = $bootFreq{$entry};
								print "\t\tBootFreq: $cladeBF\n" if ($verbose);
									$treeGrow .= "'".(sprintf "$decimalPrint", $cladeBF)."'" if ($labelType eq "both" or $labelType eq "node");
									$treeGrow .= ":".(sprintf "$decimalPrint", $cladeBF) if ($labelType eq "both" or $labelType eq "branch");
								last;
								}
							}
						if (not defined $cladeBF)	# No match: give minimal frequency
							{
							print "\t\tBootFreq: <$minFreq\n" if ($verbose);
							$treeGrow .= "'<$minFreq'" if ($labelType eq "both" or $labelType eq "node");
							$treeGrow .= ":".(sprintf "$decimalPrint", $minFreq) if ($labelType eq "both" or $labelType eq "branch");
							}
						$treeGrow .= ",";
						}
					}
				elsif (substr($tree, $position, 1) eq ",")
					{
					$treeGrow .= ",";
					}
				elsif (substr($tree, $position, 1) eq ":")	# Branch lengths present; ignore
					{
					until (substr($tree, $position+$offset, 1) eq "(" or substr($tree, $position+$offset, 1) eq ")" or substr($tree, $position+$offset, 1) eq "," or $position+$offset >= length($tree))
						{
						$offset++;
						}
					$position += $offset-1;
					}
				else	#Determine membership of clade
					{
					my $prevChar = substr($tree, $position-1, 1);
					until (substr($tree, $position+$offset, 1) eq "(" or substr($tree, $position+$offset, 1) eq ")" or substr($tree, $position+$offset, 1) eq "," or substr($tree, $position+$offset, 1) eq ":" or $position+$offset >= length($tree))
						{
						$offset++;
						}
					unless ($prevChar eq ")")	# Text represents a node label and not a taxon
						{
						for (my $clade = 1; $clade <= $CladeID; $clade++)	#Assign taxon to any open clades
							{
							if ($open[$clade])
								{
								$present{substr($tree, $position, $offset)}{$clade} = 1;
								}
							}
						$treeGrow .= substr($tree, $position, $offset);
						if ($labelType eq "both" or $labelType eq "branch")
							{
							$treeGrow .= ":100";
							if ($decimalPlace > 0)
								{
								$treeGrow .= ".";
								$treeGrow .= "0" for (1..$decimalPlace);
								}
							}
						$treeGrow .= ","
						}
					$position += $offset-1;
					}
				$position++;
				}
				$treeGrow =~ s/,+/,/g;
				$treeGrow =~ s/,\)/\)/g;
		
			print OUT "\ttree $treeName{$treeID} = ";
				if ($rootingStatus{$treeID} == 1)
					{
					print OUT "[&R]";
					}
				else
					{
					print OUT "[&U]";
					}
				print OUT " $treeGrow;\n";
			
			if ($verbose)
				{
				print "\ttree $treeName{$treeID} = ";
					if ($rootingStatus{$treeID} == 1)
						{
						print "[&R]";
						}
					else
						{
						print "[&U]";
						}
				print " $treeGrow;\n";
				}
			}
		print OUT "end;\n";
		close OUT;

# Subroutines used in the script

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}

sub treeRead	# Subroutine to read in tree descriptions and derive taxon names; accounts for any translation tables
	{
	my $treeFile = shift;
	my $treeNum = 0;

	setLineBreak($treeFile);
	open (TREES, "<$treeFile") or die "Cannot open nexus-formatted treefile $treeFile\n";
		my (@allTrees, $rawTreeDescrip, $treeDescrip, $treeFlag, $treeCount, $rooting);
		my (@taxonList, %taxonPresent);
		my (%TTable, $translation, $translateFlag, $maxTaxonID, $transText);
		
		while (<TREES>)
			{
			chomp;
			$_ =~ s/^\s+//;		# Strip (any) leading white space
			if ($_ =~ m/begin trees;/i)		# Clear variables for each tree block
				{
				undef %TTable;
				$translation = $translateFlag = $treeFlag = $maxTaxonID = $treeCount = 0;
				$treeBlock++;
				}
			if ($_ =~ m/translate/i)	#Read translation table of taxon names (if any)
				{
				$translateFlag = $translation = 1;
				$transText = "";
				print "Reading translation table ...\n" if ($verbose);
				}
			if ($translateFlag)
				{
				unless ($_ =~ m/translate/i)
					{
					$transText .= $_;
					if ($_ =~ m/;/)
						{
						$translateFlag = 0;
						$transText =~ s/;$//;	#Strip trailing semi-colon
						my @transEntries = split(/,/, $transText);
						foreach my $entry (@transEntries)
							{
							$entry =~ s/^\s+//;		#Strip (any) leading white space
							my ($taxonID, $taxonName) = split(/\s/,$entry);
							$maxTaxonID = $taxonID if ($taxonID > $maxTaxonID);
							$taxonName =~ s/\s+//g;
							$TTable{$taxonID} = $taxonName;
							}
						}
					}
				}
			if ($_ =~ /tree/i and $_ =~ /=/)
				{
				unless ($_ =~ /^>/)		#Ignore (any) commented lines
					{
					$rawTreeDescrip = $treeDescrip = "";
					unless ($block == 0 and $treeBlock > 1)
						{
						$treeFlag = 1;
						}
					}
				}
			if ($treeFlag)
				{
				$rawTreeDescrip .= $_;
				if ($_ =~ m/:/)
					{
					$SupportFlag = 1;
					}
				if ($_ =~ m/;/)
					{
					$treeCount++;
					print "\nTree block: $treeBlock\tTree number: $treeCount\n" if ($verbose);
					print "\tTranslating taxon names ...\n" if ($verbose and $translation);
					# Determine if trees are rooted or not (default for no information is not rooted)
						if ($rawTreeDescrip =~ /\[&R\]/)
							{
							$rooting = 1;
							}
						elsif ($rawTreeDescrip =~ /\[&U\]/)
							{
							$rooting = 0;
							$unrootedFlag = 1;
							}
						else
							{
							$rooting = $defaultRooting;
							$unrootedFlag = 1 if ($defaultRooting == 0);
							}
					# Remove any embedded comments
						if ($rawTreeDescrip =~ /\[/)
							{
							for (my $position = 0; $position < length($rawTreeDescrip); $position++)
								{
								if (substr($rawTreeDescrip, $position, 1) eq "[")
									{
									$position++ until (substr($rawTreeDescrip, $position, 1) eq "]");
									$position++;
									}
								else
									{
									$treeDescrip .= substr($rawTreeDescrip, $position, 1);
									}
								}
							}
						else
							{
							$treeDescrip = $rawTreeDescrip;
							}
					$treeFlag = 0;
					my $position = 0;
					until (substr($treeDescrip, $position, 1) eq "(")		# Avoid anything before tree description proper
						{
						$position++;
						}
					my $tree = substr($treeDescrip, $position);
						$tree =~ s/;\s*$//;
						$tree =~ s/\s+//g;
					my ($name) = (split /=/, $treeDescrip, 0);		# Get name of tree, if any
						$name =~ s/\s*\w*tree\s+//i;
						$name =~ s/\s*\*\s*//i;
						$name =~ s/\s+$//g;
#						$treeNames{$treeCount} = $name;
					my $openCount = ($tree =~ tr/(//);
					my $closedCount = ($tree =~ tr/)//);
					if ($openCount == $closedCount)		# Tree is balanced; process
						{
						$position = length($tree) - 1;
						while ($position >= 0)
							{
							my $IDlength = 1;
							if (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")		#Tree structure; skip to next character
								{
								$position--;
								}
							else	# Read in taxonID
								{
								until (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")
									{
									$position--;
									$IDlength++;
									}
								unless (substr($tree, $position, 1) eq ":")		# Avoid reading branch lengths as taxon names
									{
									my $readName = substr($tree, $position+1, $IDlength-1);
									if ($translation)	# Translate taxon names
										{
										if ($TTable{$readName})
											{
											substr($tree, $position+1, $IDlength-1) = $TTable{$readName};		#Translate name
											unless ($taxonPresent{$TTable{$readName}})
												{
												push @taxonList, $TTable{$readName};
												$taxonPresent{$TTable{$readName}} = 1;
												}
											}
										else
											{
											print "\t\tUnidentified taxon ".$readName." in tree #".$treeCount." of tree block #".$treeBlock.". Replaced with xxx.\n";
											substr($tree, $position+1, $IDlength-1) = "xxx";		#Translate name
											unless ($taxonPresent{"xxx"})
												{
												push @taxonList, "xxx";
												$taxonPresent{"xxx"} = 1;
												}
											}
										}
									else	#Enter name as read from tree, unless already present
										{
										unless ($taxonPresent{$readName})
											{
											push @taxonList, $readName;
											$taxonPresent{$readName} = 1;
											}
										}
									}
								}
							}
						$treeNum++;
							my $treeID = "treeID_$treeNum";
						
						push @allTrees, $treeID;
							$treeDescription{$treeID} = $tree;
							$rootingStatus{$treeID} = $rooting;
							$treeName{$treeID} = $name;
						print "\tTree description:\n\t\t$tree\n" if ($verbose);
						}
					else	#Tree is unbalanced; reject
						{
						print "\tTree #".$treeCount." of tree block #".$treeBlock." is unbalanced. Not processed.\n";
						}
					}
				}
			}
	close TREES;
	
	# Process taxon list
		@allTaxa = sort(duplStrip(@taxonList));
	
	return @allTrees;
	}

sub duplStrip	# Subroutine to strip duplicate entries from an array
	{
	my @duplArray = @_;
	my %dupl;		# Remove duplicate taxa
		my @stripArray = grep(!$dupl{$_}++, @duplArray);
		@stripArray = sort {$a cmp $b} @stripArray;
	return @stripArray;
	}

# Version history
#
#	v1.2.2 (April 22, 2005)
#		- tree names and rooting status now saved due to shifting tree reading to (more complete) subroutine
#		- minor bug fixes
#
#	v1.2.1 (March 14, 2005)
#		- fixed program so as to account for non-sequential PAUP IDs (due to taxon deletion)
#		- minor bug fixes
#
#	v1.2 (February 18, 2005)
#		- added ability to read in dot plot split over multiple lines (for big data sets)
#		- generalized script to also process majority-rule consensus output
#		- minor bug fixes
#
#	v1.1 (November 15, 2004)
#		- added option as to how to display bootstrap frequencies
#		- minor bug fixes
#
#	v1.0a (April 29, 2004)
#		- added automatic detection of line break style
#
#	v1.0 (???)
#		- initial release
